// Copyright 2014 Citra Emulator Project
// Licensed under GPLv2 or any later version
// Refer to the license.txt file included.

#include "audio_core/audio_types.h"
#include "common/assert.h"
#include "common/hash.h"
#include "common/logging/log.h"
#include "core/core.h"
#include "core/core_timing.h"
#include "core/hle/ipc_helpers.h"
#include "core/hle/kernel/process.h"
#include "core/hle/service/dsp/dsp_dsp.h"

using DspPipe = AudioCore::DspPipe;
using InterruptType = Service::DSP::DSP_DSP::InterruptType;

namespace AudioCore {
enum class DspPipe;
}

namespace Service::DSP {

class Dsp1 {
public:
    Dsp1(std::vector<u8> raw);

    struct Header {
        u8 signature[0x100];
        u8 magic[4];
        u32 binary_size;
        u16 memory_layout;
        u16 padding;
        u8 unknown;
        u8 filter_segment_type;
        u8 num_segments;
        u8 flags;
        u32 filter_segment_address;
        u32 filter_segment_size;
        u64 zero;
        struct Segment {
            u32 offset;
            u32 address;
            u32 size;
            u8 pa, pb, pc;
            u8 memory_type;
            u8 sha256[0x20];
        } segments[10];
    };

    static_assert(sizeof(Header) == 0x300, "!");

    struct Segment {
        std::vector<u8> data;
        u8 memory_type;
        u32 target;
    };

    std::vector<Segment> segments;
    bool recv_data_on_start;
};

Dsp1::Dsp1(std::vector<u8> raw) {
    Header header;
    std::memcpy(&header, raw.data(), sizeof(header));

    recv_data_on_start = (header.flags & 1) != 0;

    printf("Memory layout = %04X\n", header.memory_layout);
    printf("Unk = %02X\n", header.unknown);
    printf("Filter segment type = %d\n", header.filter_segment_type);
    printf("Num segments = %d\n", header.num_segments);
    printf("Flags = %d\n", header.flags);
    printf("Filter address = 2 * %08X\n", header.filter_segment_address);
    printf("Filter size = %08X\n", header.filter_segment_size);

    for (u32 i = 0; i < header.num_segments; ++i) {
        Segment segment;
        segment.data =
            std::vector<u8>(raw.begin() + header.segments[i].offset,
                            raw.begin() + header.segments[i].offset + header.segments[i].size);
        segment.memory_type = header.segments[i].memory_type;
        segment.target = header.segments[i].address; /*header.segments[i].address * 2 +
             (segment.memory_type == 2 ? 0x1FF40000 : 0x1FF00000);*/
        segments.push_back(segment);

        printf("[Segment %d]\n", i);
        printf("memory_type = %d\n", segment.memory_type);
        printf("target = %08X\n", segment.target);
        printf("size = %08X\n", (u32)segment.data.size());
    }
}

u16 semaphore_heartbeat = 0;
u16 pipe_base_waddr = 0;

class SemaphoreEvent : public Kernel::Event {
public:
    SemaphoreEvent() : Kernel::Event(Core::System::GetInstance().Kernel()) {
        signaled = false;
        reset_type = Kernel::ResetType::OneShot;
        name = "DSP::SemaphoreEvent";
    }

    void NotifyHLEListener() override {
        auto& teakra = Core::System::GetInstance().teakra;
        LOG_CRITICAL(Service_DSP, "heartbeat");
        teakra.SetSemaphore(semaphore_heartbeat);
    }
};

constexpr unsigned TeakraSlice = 20000;

void RunSomeTeakra() {
    auto& teakra = Core::System::GetInstance().teakra;
    teakra.Run(TeakraSlice);
}

struct PipeStatus {
    u16_le waddress;
    u16_le bsize;
    u16_le read_bptr;
    u16_le write_bptr;
    u8 slot_index;
    u8 flags;
};

static_assert(sizeof(PipeStatus) == 10, "!");

enum class PipeDirection : u8 {
    DSPtoCPU = 0,
    CPUtoDSP = 1,
};

u8* GetDspDataPointer(u32 baddr) {
    auto& teakra = Core::System::GetInstance().teakra;
    auto& memory = teakra.GetDspMemory();
    return &memory[0x40000 + baddr];
}

u8 PipeIndexToSlotIndex(u8 pipe_index, PipeDirection direction) {
    return (pipe_index << 1) + (u8)direction;
}

PipeStatus GetPipeStatus(u8 pipe_index, PipeDirection direction) {
    u8 slot_index = PipeIndexToSlotIndex(pipe_index, direction);
    PipeStatus pipe_status;
    std::memcpy(&pipe_status,
                GetDspDataPointer(pipe_base_waddr * 2 + slot_index * sizeof(PipeStatus)),
                sizeof(PipeStatus));
    ASSERT(pipe_status.slot_index == slot_index);
    return pipe_status;
}

void UpdatePipeStatus(const PipeStatus& pipe_status) {
    u8 slot_index = pipe_status.slot_index;
    u8* status_address = GetDspDataPointer(pipe_base_waddr * 2 + slot_index * sizeof(PipeStatus));
    if (slot_index % 2 == 0) {
        std::memcpy(status_address + 4, &pipe_status.read_bptr, sizeof(u16));
    } else {
        std::memcpy(status_address + 6, &pipe_status.write_bptr, sizeof(u16));
    }
}

void WritePipe(u8 pipe_index, const std::vector<u8>& data) {
    auto& teakra = Core::System::GetInstance().teakra;
    PipeStatus pipe_status = GetPipeStatus(pipe_index, PipeDirection::CPUtoDSP);
    bool need_update = false;
    const u8* buffer_ptr = data.data();
    u16 bsize = (u16)data.size();
    while (bsize != 0) {
        u16 x = pipe_status.read_bptr ^ pipe_status.write_bptr;
        ASSERT_MSG(x != 0x8000, "full");
        u16 write_bend;
        if (x > 0x8000)
            write_bend = pipe_status.read_bptr & 0x7FFF;
        else
            write_bend = pipe_status.bsize;
        u16 write_bbegin = pipe_status.write_bptr & 0x7FFF;
        ASSERT_MSG(write_bend > write_bbegin, "end {:04X} <= begin {:04X}, size {:04X}", write_bend,
                   write_bbegin, pipe_status.bsize);
        u16 write_bsize = std::min<u16>(bsize, write_bend - write_bbegin);
        std::memcpy(GetDspDataPointer(pipe_status.waddress * 2 + write_bbegin), buffer_ptr,
                    write_bsize);
        buffer_ptr += write_bsize;
        pipe_status.write_bptr += write_bsize;
        bsize -= write_bsize;
        ASSERT_MSG((pipe_status.write_bptr & 0x7FFF) <= pipe_status.bsize, "write > size");
        if ((pipe_status.write_bptr & 0x7FFF) == pipe_status.bsize) {
            pipe_status.write_bptr &= 0x8000;
            pipe_status.write_bptr ^= 0x8000;
        }
        need_update = true;
    }
    if (need_update) {
        UpdatePipeStatus(pipe_status);
        while (!teakra.SendDataIsEmpty(2))
            RunSomeTeakra();
        teakra.SendData(2, pipe_status.slot_index);
    }
}

std::vector<u8> ReadPipe(u8 pipe_index, u16 bsize) {
    auto& teakra = Core::System::GetInstance().teakra;
    PipeStatus pipe_status = GetPipeStatus(pipe_index, PipeDirection::DSPtoCPU);
    bool need_update = false;
    std::vector<u8> data(bsize);
    u8* buffer_ptr = data.data();
    while (bsize != 0) {
        u16 x = pipe_status.read_bptr ^ pipe_status.write_bptr;
        ASSERT_MSG(x != 0, "empty");
        u16 read_bend;
        if (x >= 0x8000) {
            read_bend = pipe_status.bsize;
        } else {
            read_bend = pipe_status.write_bptr & 0x7FFF;
        }
        u16 read_bbegin = pipe_status.read_bptr & 0x7FFF;
        ASSERT(read_bend > read_bbegin);
        u16 read_bsize = std::min<u16>(bsize, read_bend - read_bbegin);
        std::memcpy(buffer_ptr, GetDspDataPointer(pipe_status.waddress * 2 + read_bbegin),
                    read_bsize);
        buffer_ptr += read_bsize;
        pipe_status.read_bptr += read_bsize;
        bsize -= read_bsize;
        ASSERT_MSG((pipe_status.read_bptr & 0x7FFF) <= pipe_status.bsize, "read > size");
        if ((pipe_status.read_bptr & 0x7FFF) == pipe_status.bsize) {
            pipe_status.read_bptr &= 0x8000;
            pipe_status.read_bptr ^= 0x8000;
        }
        need_update = true;
    }
    if (need_update) {
        UpdatePipeStatus(pipe_status);
        while (!teakra.SendDataIsEmpty(2))
            RunSomeTeakra();
        teakra.SendData(2, pipe_status.slot_index);
    }
    return data;
}

u16 GetPipeReadableSize(u8 pipe_index) {
    auto& teakra = Core::System::GetInstance().teakra;
    PipeStatus pipe_status = GetPipeStatus(pipe_index, PipeDirection::DSPtoCPU);
    u16 size = pipe_status.write_bptr - pipe_status.read_bptr;
    if ((pipe_status.read_bptr ^ pipe_status.write_bptr) >= 0x8000) {
        size += pipe_status.bsize;
    }
    return size & 0x7FFF;
}

Core::TimingEventType* teakra_event;

void TeakraEvent(u64, u64 late) {
    RunSomeTeakra();
    u64 next = TeakraSlice * 2; // should *2
    if (next < late)
        next = 0;
    else
        next -= late;
    Core::System::GetInstance().CoreTiming().ScheduleEvent(next, teakra_event, 0);
}

void DSP_DSP::ProcessPipeEvent(bool event_from_data) {
    auto& teakra = Core::System::GetInstance().teakra;
    static bool semaphore_signaled = false;
    static bool data_signaled = false;
    if (event_from_data) {
        data_signaled = true;
    } else {
        if ((teakra.GetSemaphore() & 0x8000) == 0)
            return;
        semaphore_signaled = true;
    }
    if (semaphore_signaled && data_signaled) {
        semaphore_signaled = data_signaled = false;
        u16 slot = teakra.RecvData(2);
        u16 side = slot % 2;
        u16 pipe = slot / 2;
        ASSERT(pipe < 16);
        if (side != (u16)PipeDirection::DSPtoCPU)
            return;
        LOG_WARNING(Service_DSP, "Pipe signal received for pipe {}", pipe);
        if (pipe == 0) {
            Service::DSP::ReadPipe(pipe, Service::DSP::GetPipeReadableSize(pipe));
        } else {
            LOG_WARNING(Service_DSP, "Readable size = {}", Service::DSP::GetPipeReadableSize(pipe));
            GetInterruptEvent(InterruptType::Pipe, (DspPipe)pipe)->Signal();
        }
    }
}

void DSP_DSP::RecvData(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x01, 1, 0);
    const u32 register_number = rp.Pop<u32>();

    // ASSERT_MSG(register_number == 0, "Unknown register_number {}", register_number);

    // Application reads this after requesting DSP shutdown, to verify the DSP has indeed shutdown
    // or slept.

    auto& teakra = Core::System::GetInstance().teakra;
    while (!teakra.RecvDataIsReady(register_number))
        RunSomeTeakra();

    IPC::RequestBuilder rb = rp.MakeBuilder(2, 0);
    rb.Push(RESULT_SUCCESS);
    rb.Push(teakra.RecvData(register_number));
    /*switch (Core::DSP().GetDspState()) {
    case AudioCore::DspState::On:
        rb.Push<u32>(0);
        break;
    case AudioCore::DspState::Off:
    case AudioCore::DspState::Sleeping:
        rb.Push<u32>(1);
        break;
    default:
        UNREACHABLE();
        break;
    }*/

    LOG_WARNING(Service_DSP, "register_number={}", register_number);
}

void DSP_DSP::RecvDataIsReady(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x02, 1, 0);
    const u32 register_number = rp.Pop<u32>();

    // ASSERT_MSG(register_number == 0, "Unknown register_number {}", register_number);
    auto& teakra = Core::System::GetInstance().teakra;
    IPC::RequestBuilder rb = rp.MakeBuilder(2, 0);
    rb.Push(RESULT_SUCCESS);
    rb.Push(teakra.RecvDataIsReady(register_number)); /// 0 = not ready, 1 = ready to read

    LOG_WARNING(Service_DSP, "register_number={}", register_number);
}

void DSP_DSP::SetSemaphore(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x07, 1, 0);
    const u16 semaphore_value = rp.Pop<u16>();

    Core::System::GetInstance().teakra.SetSemaphore(semaphore_value);

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 0);
    rb.Push(RESULT_SUCCESS);

    LOG_WARNING(Service_DSP, "(STUBBED) called, semaphore_value={:04X}", semaphore_value);
}

void DSP_DSP::ConvertProcessAddressFromDspDram(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x0C, 1, 0);
    const u32 address = rp.Pop<u32>();

    IPC::RequestBuilder rb = rp.MakeBuilder(2, 0);
    rb.Push(RESULT_SUCCESS);

    // TODO(merry): There is a per-region offset missing in this calculation (that seems to be
    // always zero).
    rb.Push<u32>((address << 1) + (Memory::DSP_RAM_VADDR + 0x40000));

    LOG_DEBUG(Service_DSP, "address=0x{:08X}", address);
}

void DSP_DSP::WriteProcessPipe(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x0D, 2, 2);
    const u32 channel = rp.Pop<u32>();
    const u32 size = rp.Pop<u32>();
    auto buffer = rp.PopStaticBuffer();

    const DspPipe pipe = static_cast<DspPipe>(channel);
    LOG_WARNING(Service_DSP, "channel={}, size=0x{:X}, buffer_size={:X}", channel, size,
                buffer.size());
    // This behaviour was confirmed by RE.
    // The likely reason for this is that games tend to pass in garbage at these bytes
    // because they read random bytes off the stack.
    switch (pipe) {
    case DspPipe::Audio:
        ASSERT(buffer.size() >= 4);
        buffer[4] = 4;
        break;
    case DspPipe::Binary:
        ASSERT(buffer.size() >= 8);
        buffer[20] = 20;
        break;
    }

    // Core::DSP().PipeWrite(pipe, buffer);
    Service::DSP::WritePipe((u8)pipe, buffer);

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 0);
    rb.Push(RESULT_SUCCESS);
}

void DSP_DSP::ReadPipe(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x0E, 3, 0);
    const u32 channel = rp.Pop<u32>();
    const u32 peer = rp.Pop<u32>();
    const u16 size = rp.Pop<u16>();

    const DspPipe pipe = static_cast<DspPipe>(channel);
    const u16 pipe_readable_size = Service::DSP::GetPipeReadableSize((u8)pipe);

    std::vector<u8> pipe_buffer;
    if (pipe_readable_size >= size)
        pipe_buffer = Service::DSP::ReadPipe((u8)pipe, size);
    else
        UNREACHABLE(); // No more data is in pipe. Hardware hangs in this case; Should never
                       // happen.

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 2);
    rb.Push(RESULT_SUCCESS);
    rb.PushStaticBuffer(std::move(pipe_buffer), 0);

    LOG_DEBUG(Service_DSP, "channel={}, peer={}, size=0x{:04X}, pipe_readable_size=0x{:04X}",
              channel, peer, size, pipe_readable_size);
}

void DSP_DSP::GetPipeReadableSize(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x0F, 2, 0);
    const u32 channel = rp.Pop<u32>();
    const u32 peer = rp.Pop<u32>();

    const DspPipe pipe = static_cast<DspPipe>(channel);
    const u16 pipe_readable_size = Service::DSP::GetPipeReadableSize((u8)pipe);

    IPC::RequestBuilder rb = rp.MakeBuilder(2, 0);
    rb.Push(RESULT_SUCCESS);
    rb.Push<u16>(pipe_readable_size);

    LOG_DEBUG(Service_DSP, "channel={}, peer={}, return pipe_readable_size=0x{:04X}", channel, peer,
              pipe_readable_size);
}

void DSP_DSP::ReadPipeIfPossible(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x10, 3, 0);
    const u32 channel = rp.Pop<u32>();
    const u32 peer = rp.Pop<u32>();
    const u16 size = rp.Pop<u16>();

    const DspPipe pipe = static_cast<DspPipe>(channel);
    const u16 pipe_readable_size = Service::DSP::GetPipeReadableSize((u8)pipe);

    LOG_WARNING(Service_DSP, "channel={}, peer={}, size=0x{:04X}, pipe_readable_size=0x{:04X}",
                channel, peer, size, pipe_readable_size);

    std::vector<u8> pipe_buffer;
    if (pipe_readable_size >= size)
        pipe_buffer = Service::DSP::ReadPipe((u8)pipe, size);

    IPC::RequestBuilder rb = rp.MakeBuilder(2, 2);
    rb.Push(RESULT_SUCCESS);
    rb.Push<u16>(pipe_readable_size);
    rb.PushStaticBuffer(pipe_buffer, 0);
}

void DSP_DSP::LoadComponent(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x11, 3, 2);
    const u32 size = rp.Pop<u32>();
    const u32 prog_mask = rp.Pop<u32>();
    const u32 data_mask = rp.Pop<u32>();
    auto& buffer = rp.PopMappedBuffer();

    IPC::RequestBuilder rb = rp.MakeBuilder(2, 2);
    rb.Push(RESULT_SUCCESS);
    rb.Push(true); /// Pretend that we actually loaded the DSP firmware
    rb.PushMappedBuffer(buffer);

    // TODO(bunnei): Implement real DSP firmware loading

    std::vector<u8> component_data(size);
    buffer.Read(component_data.data(), 0, size);

    LOG_INFO(Service_DSP, "Firmware hash: {:#018x}",
             Common::ComputeHash64(component_data.data(), component_data.size()));
    // Some versions of the firmware have the location of DSP structures listed here.
    if (size > 0x37C) {
        LOG_INFO(Service_DSP, "Structures hash: {:#018x}",
                 Common::ComputeHash64(component_data.data() + 0x340, 60));
    }

    auto& teakra = Core::System::GetInstance().teakra;

    Dsp1 dsp(component_data);
    auto& dsp_memory = teakra.GetDspMemory();
    u8* program = dsp_memory.data();
    u8* data = dsp_memory.data() + 0x40000;
    dsp_memory.fill(0);

    for (const auto& segment : dsp.segments) {
        if (segment.memory_type == 0 || segment.memory_type == 1) {
            std::memcpy(program + segment.target * 2, segment.data.data(), segment.data.size());
        } else if (segment.memory_type == 2) {
            std::memcpy(data + segment.target * 2, segment.data.data(), segment.data.size());
        }
    }
    Core::System::GetInstance().CoreTiming().ScheduleEvent(TeakraSlice, teakra_event, 0);

    if (dsp.recv_data_on_start) {
        for (unsigned i = 0; i < 3; ++i) {
            while (!teakra.RecvDataIsReady(i))
                RunSomeTeakra();
            ASSERT(teakra.RecvData(i) == 1);
        }
    }
    LOG_INFO(Service_DSP, "DSP initialized");

    while (!teakra.RecvDataIsReady(2))
        RunSomeTeakra();
    pipe_base_waddr = teakra.RecvData(2);

    teakra.SetRecvDataHandler(
        0, [this]() { GetInterruptEvent(InterruptType::Zero, (DspPipe)0)->Signal(); });
    teakra.SetRecvDataHandler(
        1, [this]() { GetInterruptEvent(InterruptType::One, (DspPipe)0)->Signal(); });
    teakra.SetRecvDataHandler(2, [this]() { ProcessPipeEvent(true); });

    teakra.SetSemaphoreHandler([this]() { ProcessPipeEvent(false); });

    LOG_INFO(Service_DSP, "pipe_base_waddr = %04X", pipe_base_waddr);

    LOG_DEBUG(Service_DSP, "(STUBBED) called size=0x{:X}, prog_mask=0x{:08X}, data_mask=0x{:08X}",
              size, prog_mask, data_mask);
}

void DSP_DSP::FlushDataCache(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x13, 2, 2);
    const VAddr address = rp.Pop<u32>();
    const u32 size = rp.Pop<u32>();
    const auto process = rp.PopObject<Kernel::Process>();

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 0);
    rb.Push(RESULT_SUCCESS);

    LOG_TRACE(Service_DSP, "called address=0x{:08X}, size=0x{:X}, process={}", address, size,
              process->process_id);
}

void DSP_DSP::InvalidateDataCache(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x14, 2, 2);
    const VAddr address = rp.Pop<u32>();
    const u32 size = rp.Pop<u32>();
    const auto process = rp.PopObject<Kernel::Process>();

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 0);
    rb.Push(RESULT_SUCCESS);

    LOG_TRACE(Service_DSP, "called address=0x{:08X}, size=0x{:X}, process={}", address, size,
              process->process_id);
}

void DSP_DSP::RegisterInterruptEvents(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x15, 2, 2);
    const u32 interrupt = rp.Pop<u32>();
    const u32 channel = rp.Pop<u32>();
    auto event = rp.PopObject<Kernel::Event>();

    ASSERT_MSG(interrupt < NUM_INTERRUPT_TYPE && channel < AudioCore::num_dsp_pipe,
               "Invalid type or pipe: interrupt = {}, channel = {}", interrupt, channel);

    const InterruptType type = static_cast<InterruptType>(interrupt);
    const DspPipe pipe = static_cast<DspPipe>(channel);

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 0);

    if (event) { /// Register interrupt event
        if (HasTooManyEventsRegistered()) {
            LOG_INFO(Service_DSP,
                     "Ran out of space to register interrupts (Attempted to register "
                     "interrupt={}, channel={}, event={})",
                     interrupt, channel, event->GetName());
            rb.Push(ResultCode(ErrorDescription::InvalidResultValue, ErrorModule::DSP,
                               ErrorSummary::OutOfResource, ErrorLevel::Status));
            return;
        } else {
            GetInterruptEvent(type, pipe) = event;
            LOG_INFO(Service_DSP, "Registered interrupt={}, channel={}, event={}", interrupt,
                     channel, event->GetName());
        }
    } else { /// Otherwise unregister event
        GetInterruptEvent(type, pipe) = nullptr;
        LOG_INFO(Service_DSP, "Unregistered interrupt={}, channel={}", interrupt, channel);
    }
    rb.Push(RESULT_SUCCESS);
}

void DSP_DSP::GetSemaphoreEventHandle(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x16, 0, 0);

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 2);
    rb.Push(RESULT_SUCCESS);
    rb.PushCopyObjects(semaphore_event);

    LOG_DEBUG(Service_DSP, "(STUBBED) called");
}

void DSP_DSP::SetSemaphoreMask(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x17, 1, 0);
    semaphore_heartbeat = rp.Pop<u16>();

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 0);
    rb.Push(RESULT_SUCCESS);

    LOG_DEBUG(Service_DSP, "(STUBBED) called mask=0x{:08X}", semaphore_heartbeat);
}

void DSP_DSP::GetHeadphoneStatus(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x1F, 0, 0);

    IPC::RequestBuilder rb = rp.MakeBuilder(2, 0);
    rb.Push(RESULT_SUCCESS);
    rb.Push(false); /// u8, 0 = not inserted, 1 = inserted

    LOG_DEBUG(Service_DSP, "called");
}

void DSP_DSP::ForceHeadphoneOut(Kernel::HLERequestContext& ctx) {
    IPC::RequestParser rp(ctx, 0x20, 1, 0);
    const u8 force = rp.Pop<u8>();

    IPC::RequestBuilder rb = rp.MakeBuilder(1, 0);
    rb.Push(RESULT_SUCCESS);

    LOG_DEBUG(Service_DSP, "(STUBBED) called, force={}", force);
}

// DSP Interrupts:
// The audio-pipe interrupt occurs every frame tick. Userland programs normally have a thread
// that's waiting for an interrupt event. Immediately after this interrupt event, userland
// normally updates the state in the next region and increments the relevant frame counter by
// two.
void DSP_DSP::SignalInterrupt(InterruptType type, DspPipe pipe) {
    LOG_DEBUG(Service_DSP, "called, type={}, pipe={}", static_cast<u32>(type),
              static_cast<u32>(pipe));
    const auto& event = GetInterruptEvent(type, pipe);
    if (event)
        event->Signal();
}

Kernel::SharedPtr<Kernel::Event>& DSP_DSP::GetInterruptEvent(InterruptType type, DspPipe pipe) {
    switch (type) {
    case InterruptType::Zero:
        return interrupt_zero;
    case InterruptType::One:
        return interrupt_one;
    case InterruptType::Pipe: {
        const std::size_t pipe_index = static_cast<std::size_t>(pipe);
        ASSERT(pipe_index < AudioCore::num_dsp_pipe);
        return pipes[pipe_index];
    }
    }
    UNREACHABLE_MSG("Invalid interrupt type = {}", static_cast<std::size_t>(type));
}

bool DSP_DSP::HasTooManyEventsRegistered() const {
    std::size_t number =
        std::count_if(pipes.begin(), pipes.end(), [](const auto& evt) { return evt != nullptr; });

    if (interrupt_zero != nullptr)
        number++;
    if (interrupt_one != nullptr)
        number++;

    LOG_DEBUG(Service_DSP, "Number of events registered = {}", number);
    return number >= max_number_of_interrupt_events;
}

void DSP_DSP::UnloadComponent(Kernel::HLERequestContext& ctx) {
    // UNREACHABLE();
}

DSP_DSP::DSP_DSP(Core::System& system) : ServiceFramework("dsp::DSP", DefaultMaxSessions) {
    static const FunctionInfo functions[] = {
        // clang-format off
        {0x00010040, &DSP_DSP::RecvData, "RecvData"},
        {0x00020040, &DSP_DSP::RecvDataIsReady, "RecvDataIsReady"},
        {0x00030080, nullptr, "SendData"},
        {0x00040040, nullptr, "SendDataIsEmpty"},
        {0x000500C2, nullptr, "SendFifoEx"},
        {0x000600C0, nullptr, "RecvFifoEx"},
        {0x00070040, &DSP_DSP::SetSemaphore, "SetSemaphore"},
        {0x00080000, nullptr, "GetSemaphore"},
        {0x00090040, nullptr, "ClearSemaphore"},
        {0x000A0040, nullptr, "MaskSemaphore"},
        {0x000B0000, nullptr, "CheckSemaphoreRequest"},
        {0x000C0040, &DSP_DSP::ConvertProcessAddressFromDspDram, "ConvertProcessAddressFromDspDram"},
        {0x000D0082, &DSP_DSP::WriteProcessPipe, "WriteProcessPipe"},
        {0x000E00C0, &DSP_DSP::ReadPipe, "ReadPipe"},
        {0x000F0080, &DSP_DSP::GetPipeReadableSize, "GetPipeReadableSize"},
        {0x001000C0, &DSP_DSP::ReadPipeIfPossible, "ReadPipeIfPossible"},
        {0x001100C2, &DSP_DSP::LoadComponent, "LoadComponent"},
        {0x00120000, &DSP_DSP::UnloadComponent, "UnloadComponent"},
        {0x00130082, &DSP_DSP::FlushDataCache, "FlushDataCache"},
        {0x00140082, &DSP_DSP::InvalidateDataCache, "InvalidateDCache"},
        {0x00150082, &DSP_DSP::RegisterInterruptEvents, "RegisterInterruptEvents"},
        {0x00160000, &DSP_DSP::GetSemaphoreEventHandle, "GetSemaphoreEventHandle"},
        {0x00170040, &DSP_DSP::SetSemaphoreMask, "SetSemaphoreMask"},
        {0x00180040, nullptr, "GetPhysicalAddress"},
        {0x00190040, nullptr, "GetVirtualAddress"},
        {0x001A0042, nullptr, "SetIirFilterI2S1_cmd1"},
        {0x001B0042, nullptr, "SetIirFilterI2S1_cmd2"},
        {0x001C0082, nullptr, "SetIirFilterEQ"},
        {0x001D00C0, nullptr, "ReadMultiEx_SPI2"},
        {0x001E00C2, nullptr, "WriteMultiEx_SPI2"},
        {0x001F0000, &DSP_DSP::GetHeadphoneStatus, "GetHeadphoneStatus"},
        {0x00200040, &DSP_DSP::ForceHeadphoneOut, "ForceHeadphoneOut"},
        {0x00210000, nullptr, "GetIsDspOccupied"},
        // clang-format on
    };

    RegisterHandlers(functions);

    semaphore_event = Kernel::SharedPtr<Kernel::Event>(new SemaphoreEvent);
    teakra_event = system.CoreTiming().RegisterEvent("teakra", TeakraEvent);
}

DSP_DSP::~DSP_DSP() {
    semaphore_event = nullptr;
    pipes = {};
}

void InstallInterfaces(Core::System& system) {
    auto& service_manager = system.ServiceManager();
    auto dsp = std::make_shared<DSP_DSP>(system);
    dsp->InstallAsService(service_manager);
    // Core::DSP().SetServiceToInterrupt(std::move(dsp));
}

} // namespace Service::DSP
